@@ -22,10 +22,9 @@
 
 #include <R.h>
 #include <Rinternals.h>
+#include <stdbool.h>
 
 SEXP read_nbt(SEXP r_rawval, SEXP r_max_elements);
-SEXP read_nbt_compound_payload(const unsigned char** ptr, const unsigned char* end,
-        int max_elements);
-
+SEXP read_nbt_values(const unsigned char** ptr, const unsigned char* end, int max_elements, bool stop_on_end);
 SEXP write_nbt(SEXP r_value);
 
