@@ -40,3 +40,8 @@ bool scalar_logical(SEXP x);
 size_t scalar_size(SEXP x);
 int scalar_int(SEXP x);
 const char *scalar_character(SEXP x);
+
+SEXP create_stretchy_list();
+void grow_stretchy_list(SEXP l, SEXP s);
+SEXP create_stretchy_list_with_name(SEXP s, SEXP tag);
+void grow_stretchy_list_with_name(SEXP l, SEXP s, SEXP tag);
